.\" file MASS/mvrnorm.d
.\" copyright (C) 1994-9 W. N. Venables and B. D. Ripley
.\"
.BG
.FN mvrnorm
.TL
Simulate from a Multivariate Normal Distribution
.DN
Produces one or more samples from the specified
multivariate normal distribution.
.CS
mvrnorm(n=1, mu, Sigma, tol = 1e-6, empirical = F)
.RA
.AG n
the number of samples required.
.OA
.AG mu
a vector giving the means of the variables.
.AG Sigma
a positive-definite symmetric matrix specifying the covariance matrix
of the variables.
.AG tol
tolerance (relative to largest variance) for numerical lack of
positive-definiteness in `Sigma'.
.AG empirical
logical. If true, mu and Sigma specify the empirical not population
mean and covariance matrix.
.RT
If `n = 1' a vector of the same length as `mu', otherwise an
`n' by `length(mu)' matrix with one sample in each row.
.SE
Causes creation of the dataset `.Random.seed' if it does
not already exist, otherwise its value is updated.
.DT
The matrix decomposition is done via `eigen'; although a Choleski
decomposition might be faster, the eigendecomposition is 
stabler.
.SH REFERENCE
B. D. Ripley (1987)
.ul
Stochastic Simulation.
Wiley. Page 98.
.SA
`rnorm', `rmvnorm' in S-PLUS 4.0 and later.
.EX
> Sigma <- matrix(c(10,3,3,2),2,2)
> Sigma
     [,1] [,2] 
[1,]   10    3
[2,]    3    2
> var(mvrnorm(n=1000, rep(0, 2), Sigma))
         [,1]     [,2] 
[1,] 9.906492 3.030130
[2,] 3.030130 2.020929
.KW distribution
.KW multivariate
.WR
